//-------------------------------------------------------------------
//!w֌Wwb_
//!@memo 08/07/18 mmm
//-------------------------------------------------------------------

#ifndef _MMM_MATH_H_
#define _MMM_MATH_H_

#include <cmath>
#include <cassert>

#ifdef MMM_USE_D3DX
	struct D3DXVECTOR3;
	struct D3DXMATRIX;
#endif

namespace mmm
{
	template<typename T> void Swap(T* l, T* r)
	{
		T t = *l;
		*l = *r;
		*r = t;
	}
	template<typename T> T Max(T l, T r)
	{
		return (l > r)? l : r;
	}
	template<typename T> T Min(T l, T r)
	{
		return (l < r)? l : r;
	}
}
namespace mmm
{
	static float NEAR_ZERO_THD = (0.00000001f);
	static float NEAR_ONE_THD = (0.00000001f);

	class Vector3;
	class Matrix;

	//-------------------------------------------------------
	//!l߂̔֐
	//-------------------------------------------------------
	inline bool IsNear(float v, float hope, float range)
	{
		return fabs(v - hope) <= range;
	}

	//-------------------------------------------------------
	//!R̃xNgNX
	//-------------------------------------------------------
	class Vector3
	{
	public:
		Vector3(){}
		Vector3(float val) : x_(val), y_(val), z_(val){}
		Vector3(float x, float y, float z) : x_(x), y_(y), z_(z){}
		template<typename T> Vector3(T x, T y, T z) : x_(static_cast<float>(x)), y_(static_cast<float>(y)), z_(static_cast<float>(z)){}
		Vector3(const Vector3& v) : x_(v.x_), y_(v.y_), z_(v.z_){}
	public:
		#define DECLARE(ope) x_ ope rhs.x_; y_ ope rhs.y_; z_ ope rhs.z_;

		Vector3& operator  =(const Vector3& rhs){ DECLARE(=); return *this; }
		Vector3& operator +=(const Vector3& rhs){ DECLARE(+=); return *this; }
		Vector3& operator -=(const Vector3& rhs){ DECLARE(-=); return *this; }

		#undef DECLARE
	public:
		Vector3& operator *=(float rhs){ x_ *= rhs; y_ *= rhs; z_ *= rhs; return *this; }
		Vector3& operator /=(float rhs)
		{
			if(rhs == 0.f) return *this;
			float inv = 1.f / rhs;
			x_ *= inv; y_ *= inv; z_ *= inv;
			return *this;
		}
		friend Vector3 operator +(const Vector3& lhs, const Vector3& rhs){ return Vector3(lhs.x_ + rhs.x_, lhs.y_ + rhs.y_, lhs.z_ + rhs.z_); }
		friend Vector3 operator -(const Vector3& lhs, const Vector3& rhs){ return Vector3(lhs.x_ - rhs.x_, lhs.y_ - rhs.y_, lhs.z_ - rhs.z_); }
		friend Vector3 operator *(const Vector3& lhs, float rhs){ return Vector3(lhs.x_ * rhs, lhs.y_ * rhs, lhs.z_ * rhs); }
		friend Vector3 operator /(const Vector3& lhs, float rhs)
		{
			if(rhs == 0.f) return lhs;
			float inv = 1.f / rhs;
			return Vector3(lhs.x_ * inv, lhs.y_ * inv, lhs.z_ * inv);
		}
		friend Vector3 operator -(const Vector3& v)
		{
			return Vector3(-v.x_, -v.y_, -v.z_);
		}
	public:
		bool operator ==(const Vector3& rhs) const { return x_ == rhs.x_ && y_ == rhs.y_ && z_ == rhs.z_; }
		bool operator !=(const Vector3& rhs) const { return !((*this) == rhs); }
	public:
		void Set(float val){ x_ = y_ = z_ = val; }
		void Set(float x, float y, float z){ x_ = x; y_ = y; z_ = z; }
		void SetInt(int x, int y, int z){ x_ = (float)x; y_ = (float)y; z_ = (float)z; }
		void Set(const Vector3& v){ *this = v; }
		float Length() const { return sqrt(x_*x_+y_*y_+z_*z_); }
		float LengthSq() const { return x_*x_+y_*y_+z_*z_; }
	public:
		//!K֐
		void Normalize()
		{
			float len = Length();
			if(len == 0.f) return;
			if(IsNear(len, 1.f, NEAR_ONE_THD)) return;

			float inv = 1.f / len;
			x_ *= inv; y_ *= inv; z_ *= inv;
		}
		//!傫ꍇ̐K֐
		void NormalizeKnowLength(float len)
		{
			if(len == 0.f) return;
			if(IsNear(len, 1.f, NEAR_ONE_THD)) return;

			float inv = 1.f / len;
			x_ *= inv; y_ *= inv; z_ *= inv;
		}
		//!傫̐֐
		void Truncate(float max)
		{
			float len = Length();
			if(len >= max)
			{
				NormalizeKnowLength(len);
				*this *= max;
			}
		}
		//!ΒlƂčł傫vf̔ԍԂ֐
		int AbsMaxParamIndex() const
		{
			int id = 0;
			float max = fabs(p_[0]);
			for(int i = 1; i < ID_COUNT; ++ i)
			{
				float temp = fabs(p_[i]); 
				if(max < temp)
				{
					id = i;
					max = temp;
				}
			}

			return id;
		}
		//!ΒlƂčłvf̔ԍԂ֐
		int AbsMinParamIndex() const
		{
			int id = 0;
			float min = fabs(p_[0]);
			for(int i = 1; i < ID_COUNT; ++ i)
			{
				float temp = fabs(p_[i]); 
				if(min > temp)
				{
					id = i;
					min = temp;
				}
			}

			return id;
		}
	public:
	#ifdef MMM_USE_D3DX
		D3DXVECTOR3& ToD3D(){ return (D3DXVECTOR3&)(*this); }
		const D3DXVECTOR3& ToD3D() const { return (const D3DXVECTOR3&)(*this); }
	#endif
	public:
		float X() const { return x_; }
		float Y() const { return y_; }
		float Z() const { return z_; }
	public:
		operator float* (){ return (float*)this; }
		operator const float* (){ return (const float*)this; }
	public:
		enum
		{
			ID_X, ID_Y, ID_Z,
			ID_COUNT
		};
	public:
		//f[^o
		union 
		{
			float p_[ID_COUNT];
			struct
			{
				float x_, y_, z_;
			};
		};
	public:
		static const Vector3 AXIS[ID_COUNT];	//!<xNg
	};

	inline float Dot(const Vector3& lhs, const Vector3& rhs)
	{
		return lhs.x_*rhs.x_ + lhs.y_*rhs.y_ + lhs.z_*rhs.z_;
	}
	inline float DotAbs(const Vector3& lhs, const Vector3& rhs)
	{
		return fabs(lhs.x_*rhs.x_ + lhs.y_*rhs.y_ + lhs.z_*rhs.z_);
	}
	inline Vector3* Cross(Vector3* out, const Vector3& lhs, const Vector3& rhs)
	{
		out->Set(lhs.z_*rhs.x_ - lhs.x_*rhs.z_, lhs.y_*rhs.z_ - lhs.z_*rhs.y_, lhs.x_*rhs.y_ - lhs.y_*rhs.x_);
		return out;
	}

	//-----------------------------------------------------------
	//! sNX
	//-----------------------------------------------------------
	class Matrix
	{
	public:
		Matrix(){}
		Matrix(	float _11, float _12, float _13, float _14,
				float _21, float _22, float _23, float _24,
				float _31, float _32, float _33, float _34,
				float _41, float _42, float _43, float _44);
		Matrix(const Matrix& rhs);
	public:
		Matrix operator = (const Matrix& rhs);
		Matrix operator + (const Matrix& rhs) const;
		Matrix operator - (const Matrix& rhs) const;
		Matrix operator * (const Matrix& rhs) const;
	public:
		Matrix& operator +=(const Matrix& rhs);
		Matrix& operator -=(const Matrix& rhs);
		Matrix& operator *=(const Matrix& rhs);
	public:
		Matrix operator * (float s) const;
		Matrix operator / (float s) const;
		Matrix& operator *=(float s);
		Matrix& operator /=(float s);
		float* operator [](int id){ return m_[id]; }
	public:
		void SetVector(int i, const Vector3& v);
	public:
//		void Zero();
		void Identity();
		Matrix* Inverse(Matrix* pOut);
		Matrix* Scaling(const Vector3& v);
		Matrix* Rotation(const Vector3& v);
		Matrix* RotationYaw(float rad);
		Matrix* RotationPitch(float rad);
		Matrix* RotationRoll(float rad);
		Matrix* Translation(const Vector3& v);
		Matrix* Translation(float x, float y, float z);
		Matrix* ProjectionMatrix(float nearClip, float farClip, float fov, float aspect);
		Matrix* ViewMatrix(const Vector3& from, const Vector3& at, const Vector3& wup);
		Matrix* ViewportMatrix(float x, float y, float width, float height, float minZ, float maxZ);
		Matrix* ViewToTextureMatrix(float w, float h);
		Vector3* GetVectorX(Vector3* pOut);
		Vector3* GetVectorY(Vector3* pOut);
		Vector3* GetVectorZ(Vector3* pOut);
	public:
	#ifdef MMM_USE_D3DX
		D3DXMATRIX& ToD3D(){ return (D3DXMATRIX&)*this; }
		const D3DXMATRIX& ToD3D() const { return (const D3DXMATRIX&)(*this); }
#endif
	public:
		enum{ COUNT = 4 * 4 };
	public:
		union 
		{
			float m_[4][4];
			struct
			{
				float _11, _12, _13, _14;
				float _21, _22, _23, _24;
				float _31, _32, _33, _34;
				float _41, _42, _43, _44;
			};
		};
	public:
		static const Matrix E;	//Pʍs
	};
}

#endif
