//-------------------------------------------------------------------
//!_\[X
//!@memo 08/07/26 mmm
//-------------------------------------------------------------------

#include "MMMMath.h"
#include "MMMVertex.h"

namespace mmm
{
//-------------------------------------------------------------------
//_錾̂܂Ƃ߃NX
//-------------------------------------------------------------------
	VertexDecManager::VertexDecManager()
	{
		memset(decs_, 0, sizeof(decs_));
	}
	VertexDecManager::~VertexDecManager()
	{
		ReleaseDeclareVertexTypes();
	}
	//!_錾̏
	bool VertexDecManager::CreateDeclareVertexTypes(LPDIRECT3DDEVICE9 pDev)
	{
		D3DVERTEXELEMENT9 decls[VERTEX_TYPE_MAX][8] =
		{
			{
				{0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
				{0, sizeof(float)*3, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0},
				{1, 0, D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR, 0},
				D3DDECL_END()
			},
			{
				{0, 0, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITIONT, 0},
				{0, sizeof(float)*4, D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR, 0},
				D3DDECL_END()
			},
		};

		for(int i = 0; i < VERTEX_TYPE_MAX; ++ i)
		{
			//_錾̐
			if(FAILED(pDev->CreateVertexDeclaration(decls[i], &decs_[i].pDec_)))
			{
				return false;
			}

			//_錾eueAXgCh߂Ă
			UINT stride = 0;
			UINT fvf = 0;
			UINT texCnt = 0;
			for(int j = 0; decls[i][j].Type != D3DDECLTYPE_UNUSED; ++ j)
			{
				switch(decls[i][j].Type)
				{
				case D3DDECLTYPE_FLOAT1:
				case D3DDECLTYPE_FLOAT2:
				case D3DDECLTYPE_FLOAT3:
				case D3DDECLTYPE_FLOAT4: stride += sizeof(float) * (decls[i][j].Type - D3DDECLTYPE_FLOAT1+1); break;
				case D3DDECLTYPE_D3DCOLOR: stride += sizeof(DWORD); break;
				default: MMM_HALT("invalid type."); break;
				};

				switch(decls[i][j].Usage)
				{
				case D3DDECLUSAGE_POSITION: fvf |= D3DFVF_XYZ; break;
				case D3DDECLUSAGE_POSITIONT: fvf |= D3DFVF_XYZRHW; break;
				case D3DDECLUSAGE_NORMAL: fvf |= D3DFVF_NORMAL; break;
				case D3DDECLUSAGE_COLOR: fvf |= D3DFVF_DIFFUSE; break;
				case D3DDECLUSAGE_TEXCOORD: fvf |= (++ texCnt) << D3DFVF_TEXCOUNT_SHIFT; break;
//				case D3DDECLUSAGE_BLENDWEIGHT: break;
//				case D3DDECLUSAGE_BLENDINDICES: break;
				default: MMM_HALT("invalid type."); break;
				};
			}
			decs_[i].stride_ = stride;
			decs_[i].fvf_ = fvf;
		}

		return true;
	}
	//!_錾̉֐
	bool VertexDecManager::ReleaseDeclareVertexTypes()
	{
		for(int i = 0; i < VERTEX_TYPE_MAX; ++ i)
		{
			SafeReleaseD3D(decs_[i].pDec_);
			decs_[i].stride_ = 0;
			decs_[i].fvf_ = 0;
		}

		return true;
	}

//-------------------------------------------------------------------
//WIgobt@NX
//-------------------------------------------------------------------
	//!_obt@̍쐬֐
	bool GeometryBuffer::CreateVertex(LPDIRECT3DDEVICE9 pDev, UINT count, int type, void* pMem)
	{
		UINT stride = GetVertexDecManager().GetStride(type);
		UINT size = stride * count;
		UINT fvf = GetVertexDecManager().GetFVF(type);
		if(FAILED(pDev->CreateVertexBuffer(size, 0, fvf, D3DPOOL_MANAGED, &pVertex_, 0)))
		{
			return false;
		}

		vertexType_ = type;
		vertexCount_ = count;

		if(pMem)
		{
			_SetVertexBuffer(pMem, 0, size);
		}

		return true;
	}
	bool GeometryBuffer::CreateIndecies(LPDIRECT3DDEVICE9 pDev, UINT count, void* pMem)
	{
		UINT size = count * MMM_INDEX_FMT_SIZE;
		if(FAILED(pDev->CreateIndexBuffer(size, 0, MMM_INDEX_FMT, D3DPOOL_MANAGED, &pIndex_, 0)))
		{
			return false;
		}

		if(pMem)
		{
			_SetIndexBuffer(pMem, 0, size);
		}

		return true;
	}
	bool GeometryBuffer::Release()
	{
		SafeReleaseD3D(pVertex_);
		SafeReleaseD3D(pIndex_);
		vertexType_ = 0;
		vertexCount_ = 0;
		indexCount_ = 0;

		return true;
	}
	//!_obt@ւ̃f[^̐ݒ֐
	void GeometryBuffer::_SetVertexBuffer(void* pMem, UINT offset, UINT size)
	{
		void* pVs = 0;
		pVertex_->Lock(offset, size, &pVs, 0);
		memcpy(pVs, pMem, size);
		pVertex_->Unlock();
	}
	//!CfbNXobt@ւ̃f[^̐ݒ֐
	void GeometryBuffer::_SetIndexBuffer(void* pMem, UINT offset, UINT size)
	{
		void* pVs = 0;
		pIndex_->Lock(offset, size, &pVs, 0);
		memcpy(pVs, pMem, size);
		pIndex_->Unlock();
	}
	bool GeometryBuffer::SetVertex3D(void* pMem)
	{
		if(pVertex_)
		{
			UINT size = vertexCount_ * sizeof(Vertex3D);
			UINT offset = GetVertex3DOffset();

			_SetVertexBuffer(pMem, offset, size);

			return true;
		}
		return false;
	}
	bool GeometryBuffer::SetVertexColor(void* pMem)
	{
		if(pVertex_)
		{
			UINT size = vertexCount_ * sizeof(VertexColor);
			UINT offset = GetVertexColorOffset();

			_SetVertexBuffer(pMem, offset, size);

			return true;
		}
		return false;
	}
	bool GeometryBuffer::SetIndecies(void* pMem)
	{
		if(pIndex_)
		{
			UINT size = indexCount_ * MMM_INDEX_FMT_SIZE;
			_SetIndexBuffer(pMem, 0, size);
		}

		return true;
	}
	//!`p̐ݒ֐
	void GeometryBuffer::SettingRenderSource()
	{
		if(pVertex_)
		{
			LPDIRECT3DDEVICE9 pDev = 0;
			pVertex_->GetDevice(&pDev);
			if(pDev)
			{
				switch(vertexType_)
				{
				case VERTEX_TYPE_3DBASE:
					pDev->SetVertexDeclaration(GetVertexDecManager().GetDeclare(vertexType_));
					pDev->SetStreamSource(0, pVertex_, GetVertex3DOffset(), sizeof(Vertex3D));
					pDev->SetStreamSource(1, pVertex_, GetVertexColorOffset(), sizeof(VertexColor));
				break;
				case VERTEX_TYPE_2DBASE:
					pDev->SetVertexDeclaration(GetVertexDecManager().GetDeclare(vertexType_));
					pDev->SetStreamSource(0, pVertex_, GetVertex2DOffset(), sizeof(Vertex2D));
					pDev->SetStreamSource(1, pVertex_, GetVertexColorOffset(), sizeof(VertexColor));
				break;
				default:
				break;
				};

				pDev->Release();
			}
		}
	}
}
