//-------------------------------------------------------------------
//!_wb_
//!@memo 08/07/26 mmm
//-------------------------------------------------------------------

#ifndef __MMM_VERTEX_H__
#define __MMM_VERTEX_H__

#include "MMMMath.h"
#include "MMMGraphic.h"

//-------------------------------------------------------------------
//CfbNXf[^̌^Ɋւݒ}N
//-------------------------------------------------------------------
#define MMM_INDEX_FMT		D3DFMT_INDEX16
#define MMM_INDEX_FMT_SIZE	sizeof(WORD)

namespace mmm
{
	//-------------------------------------------------------------------
	//!RcpʒuA@_
	//-------------------------------------------------------------------
	struct Vertex3D
	{
		Vector3 pos_;
		Vector3 n_;
	};
	//-------------------------------------------------------------------
	//!QcpʒuA_F_
	//-------------------------------------------------------------------
	struct Vertex2D
	{
		Vector3 pos_;
		float w_;
	};
	//-------------------------------------------------------------------
	//!_F_
	//-------------------------------------------------------------------
	struct VertexColor
	{
		DWORD color_;
	};
	//-------------------------------------------------------------------
	//!eNX`f[^_
	//-------------------------------------------------------------------
	struct VertexTex
	{
		float u_, v_;
	};
	//-------------------------------------------------------------------
	//!_̃^Cvԍ
	//-------------------------------------------------------------------
	enum
	{
		VERTEX_TYPE_3DBASE,
		VERTEX_TYPE_2DBASE,
		VERTEX_TYPE_MAX
	};
	//-------------------------------------------------------------------
	//!_錾̂܂Ƃ߃NX
	//-------------------------------------------------------------------
	class VertexDecManager
	{
	private:
		VertexDecManager();
		~VertexDecManager();
	public:
		#define GetVertexDecManager() VertexDecManager::GetInstance()
		static VertexDecManager& GetInstance()
		{
			static VertexDecManager inst_;
			return inst_;
		}
	public:
		bool CreateDeclareVertexTypes(LPDIRECT3DDEVICE9 pDev);
		bool ReleaseDeclareVertexTypes();
	public:
		LPDIRECT3DVERTEXDECLARATION9 GetDeclare(int type) const
		{
			MMM_CHECK_INDEX(type, 0, VERTEX_TYPE_MAX);
			return decs_[type].pDec_;
		}
		UINT GetStride(int type) const
		{
			MMM_CHECK_INDEX(type, 0, VERTEX_TYPE_MAX-1);
			return decs_[type].stride_;
		}
		UINT GetFVF(int type) const
		{
			MMM_CHECK_INDEX(type, 0, VERTEX_TYPE_MAX-1);
			return decs_[type].fvf_;
		}
		bool Is2D(int type) const
		{
			MMM_CHECK_INDEX(type, 0, VERTEX_TYPE_MAX-1);
			return (decs_[type].fvf_ & D3DFVF_XYZRHW) > 0;
		}
		bool IsUseColor(int type) const
		{
			MMM_CHECK_INDEX(type, 0, VERTEX_TYPE_MAX-1);
			return (decs_[type].fvf_ & D3DFVF_DIFFUSE) > 0;
		}
	private:
		struct DecData
		{
			LPDIRECT3DVERTEXDECLARATION9 pDec_;
			UINT fvf_;
			UINT stride_;
		};
	private:
		DecData decs_[VERTEX_TYPE_MAX];
	};
	//-------------------------------------------------------------------
	//!WIgobt@NX
	//-------------------------------------------------------------------
	class GeometryBuffer
	{
	public:
		GeometryBuffer() : vertexType_(0), vertexCount_(0), pVertex_(0), pIndex_(0){}
		~GeometryBuffer(){ MMM_CHECK(pVertex_ == 0); MMM_CHECK(pIndex_ == 0); }
	public:
		bool CreateVertex(LPDIRECT3DDEVICE9 pDev, UINT count, int type, void* pMem = 0);
		bool CreateIndecies(LPDIRECT3DDEVICE9 pDev, UINT count, void* pMem = 0);
		bool Release();
	private:
		void _SetVertexBuffer(void* pMem, UINT offset, UINT size);
		void _SetIndexBuffer(void* pMem, UINT offset, UINT size);
	public:
		bool SetVertex3D(void* pMem);
		bool SetVertexColor(void* pMem);
		bool SetIndecies(void* pMem);
	public:
		void SettingRenderSource();
	private:
		bool IsUse3D() const { return GetVertexDecManager().Is2D(vertexType_) == false; }
		bool IsUseColor() const { return GetVertexDecManager().IsUseColor(vertexType_); }
		UINT GetVertex3DOffset() const { return 0; }
		UINT GetVertex2DOffset() const { return 0; }
		UINT GetVertexColorOffset() const { return ((IsUse3D())? sizeof(Vertex3D) : sizeof(Vertex2D)) * vertexCount_; }
		UINT GetVertexTexOffset() const {return GetVertexColorOffset() + IsUseColor() * sizeof(VertexColor) * vertexCount_;}
	private:
		int vertexType_;					//!<_̎
		UINT vertexCount_;					//!<_̐
		IDirect3DVertexBuffer9* pVertex_;	//!<_obt@
		UINT indexCount_;					//!<CfbNX̐
		IDirect3DIndexBuffer9* pIndex_;		//!<CfbNXobt@
	};
}

#endif